window.pdfjsLib.GlobalWorkerOptions.workerSrc = '/printer_bridge/static/lib/js/pdf.worker.js';

// Tools
openerp.pdfJs = window.pdfjsLib || _.noop();
openerp.printJs = window.printJS || _.noop();
openerp.html2canvas = window.html2canvas || _.noop();
openerp.html2pdf = window.html2pdf || _.noop();

/**
 * 
 * @param {*} instance 
 * @param {*} local 
 */
openerp.printer_bridge = function (instance, local) {

    // Clase que maneja el socket de conexión entre odoo y el programa de impresión cliente encargada de la impresión directa
    var SocketManager = instance.web.Class.extend({
        init: function () {
            this.socket = null;
            this.start();
        },
        start: function () {
            this.open_socket();
        },
        open_socket: function () {
            var self = this;
            var setup = self.setup_socket = function (config) {
                self.config = config;
                self.config.is_mobile = self.is_mobile_browser()

                if (!self.config.is_mobile) {
                    self.socket = io(location.protocol + '//' + config.host + ':' + config.port, {
                        path: config.path
                    });
                    self.socket.on('connect', self.handle_connect);
                    self.socket.on('error', self.handle_error);
                    self.socket.on('connect_error', self.handle_connect_error);
                    self.socket.on('request_printer_name', self.handle_printer_selection);
                    self.socket.on('show_print_status', self.handle_print_status);
                    self.socket.on('download_data', self.handle_download_data);
                }
            }

            this.get_socket_config().then(setup);
        },
        update_config: function () {
            instance.web.blockUI();
            var self = this;

            this.get_socket_config().then(function (config) {
                instance.web.unblockUI();
                self.config = _.extend(self.config, config);

                self.socket.destroy();
                self.setup_socket(config);
            });
        },
        handle_connect: function () {
            $('#printer-status').removeClass();
            $('#printer-status').addClass('printer-status-online');
        },
        handle_error: function (e) {
            instance.webclient.crashmanager.show_message(e);
        },
        handle_connect_error: function (e) {
            $('#printer-status').removeClass();
            $('#printer-status').addClass('printer-status-offline');
        },
        handle_printer_selection: function (data) {
            instance.web.unblockUI();
            var self = this;

            if (data.printers && data.printers.length === 0) {
                instance.web.notification.do_notify('Impresión', 'No hay impresoras instaladas');
                return;
            }

            var widget = new PrinterSelectionDialog(data);
            widget.get_selection().then(function (request) {
                self.emit('request_print', request);
            });
        },
        handle_print_status: function (data) {
            instance.web.unblockUI();

            if (_.isEqual(data.status, 'failed')) {
                instance.web.notification.do_warn('Impresión', 'Ha fallado la cola de impresión');
                return;
            }

            if (_.isEqual(data.status, 'canceled')) {
                instance.web.notification.do_notify('Impresión', 'Se ha cancelado la impresión');
                return;
            }

            if (_.isEqual(data.status, 'unknown')) {
                instance.web.notification.do_warn('Impresión', 'Imposible imprimir un formato de datos desconocido');
                return;
            }

            if (_.isEqual(data.status, 'printing')) {
                instance.web.notification.do_notify('Impresión', 'La impresora ' + data.printer + ' está imprimiendo');
                return;
            }

            if (_.isEqual(data.status, 'printed')) {
                instance.web.notification.do_notify('Impresión', 'La impresora ' + data.printer + ' finalizó la impresión');
                return;
            }

            if (_.isEqual(data.status, 'error')) {
                instance.web.notification.do_warn('Impresión', 'Ocurrió un error al imprimir: ' + data.printer, data.cause);
            }
        },
        handle_download_data: function (data) {
            instance.web.blockUI();

            instance.pdfJs.getDocument({
                data: atob(data)
            }).then(function (pdf) {
                if (local.socket_manager.config.action_download_pdf) {
                    pdf.getData().then(function (data) {
                        instance.web.unblockUI();
    
                        var url = instance.pdfJs.createObjectURL(data, 'application/pdf');
                        var a = document.createElement('a');
    
                        a.href = url;
                        a.download = 'download.pdf';
                        a.target = '_parent';
                        a.click();
                    });
                    return;
                }

                preview_pdf(null, pdf);
            });
        },
        get_socket_config: function () {
            var url = '/printer_bridge/socket_config';
            return this.get_from_server(url);
        },
        get_from_server: function (url) {
            instance.web.blockUI();
            return $.Deferred(function (d) {
                instance.web.jsonRpc(url, 'call').done(function () {
                    instance.web.unblockUI();
                    d.resolve.apply(d, arguments);
                }).fail(function (e) {
                    instance.web.unblockUI();
                    d.reject.apply(d, arguments);
                });
            });
        },
        is_mobile_browser: function () {
            var userAgent = navigator.userAgent;
            return /(android|bb\d+|meego).+mobile|avantgo|bada\/|blackberry|blazer|compal|elaine|fennec|hiptop|iemobile|ip(hone|od)|iris|kindle|lge |maemo|midp|mmp|mobile.+firefox|netfront|opera m(ob|in)i|palm( os)?|phone|p(ixi|re)\/|plucker|pocket|psp|series(4|6)0|symbian|treo|up\.(browser|link)|vodafone|wap|windows ce|xda|xiino/i.test(userAgent) || /1207|6310|6590|3gso|4thp|50[1-6]i|770s|802s|a wa|abac|ac(er|oo|s\-)|ai(ko|rn)|al(av|ca|co)|amoi|an(ex|ny|yw)|aptu|ar(ch|go)|as(te|us)|attw|au(di|\-m|r |s )|avan|be(ck|ll|nq)|bi(lb|rd)|bl(ac|az)|br(e|v)w|bumb|bw\-(n|u)|c55\/|capi|ccwa|cdm\-|cell|chtm|cldc|cmd\-|co(mp|nd)|craw|da(it|ll|ng)|dbte|dc\-s|devi|dica|dmob|do(c|p)o|ds(12|\-d)|el(49|ai)|em(l2|ul)|er(ic|k0)|esl8|ez([4-7]0|os|wa|ze)|fetc|fly(\-|_)|g1 u|g560|gene|gf\-5|g\-mo|go(\.w|od)|gr(ad|un)|haie|hcit|hd\-(m|p|t)|hei\-|hi(pt|ta)|hp( i|ip)|hs\-c|ht(c(\-| |_|a|g|p|s|t)|tp)|hu(aw|tc)|i\-(20|go|ma)|i230|iac( |\-|\/)|ibro|idea|ig01|ikom|im1k|inno|ipaq|iris|ja(t|v)a|jbro|jemu|jigs|kddi|keji|kgt( |\/)|klon|kpt |kwc\-|kyo(c|k)|le(no|xi)|lg( g|\/(k|l|u)|50|54|\-[a-w])|libw|lynx|m1\-w|m3ga|m50\/|ma(te|ui|xo)|mc(01|21|ca)|m\-cr|me(rc|ri)|mi(o8|oa|ts)|mmef|mo(01|02|bi|de|do|t(\-| |o|v)|zz)|mt(50|p1|v )|mwbp|mywa|n10[0-2]|n20[2-3]|n30(0|2)|n50(0|2|5)|n7(0(0|1)|10)|ne((c|m)\-|on|tf|wf|wg|wt)|nok(6|i)|nzph|o2im|op(ti|wv)|oran|owg1|p800|pan(a|d|t)|pdxg|pg(13|\-([1-8]|c))|phil|pire|pl(ay|uc)|pn\-2|po(ck|rt|se)|prox|psio|pt\-g|qa\-a|qc(07|12|21|32|60|\-[2-7]|i\-)|qtek|r380|r600|raks|rim9|ro(ve|zo)|s55\/|sa(ge|ma|mm|ms|ny|va)|sc(01|h\-|oo|p\-)|sdk\/|se(c(\-|0|1)|47|mc|nd|ri)|sgh\-|shar|sie(\-|m)|sk\-0|sl(45|id)|sm(al|ar|b3|it|t5)|so(ft|ny)|sp(01|h\-|v\-|v )|sy(01|mb)|t2(18|50)|t6(00|10|18)|ta(gt|lk)|tcl\-|tdg\-|tel(i|m)|tim\-|t\-mo|to(pl|sh)|ts(70|m\-|m3|m5)|tx\-9|up(\.b|g1|si)|utst|v400|v750|veri|vi(rg|te)|vk(40|5[0-3]|\-v)|vm40|voda|vulc|vx(52|53|60|61|70|80|81|83|85|98)|w3c(\-| )|webc|whit|wi(g |nc|nw)|wmlb|wonu|x700|yas\-|your|zeto|zte\-/i.test(userAgent.substr(0, 4));
        },
        request_print: function (data, print_directly) {
            if (!data) {
                instance.web.notification.do_warn('Impresión', 'No hay datos para imprimir');
                return;
            }

            if (!this.socket || !this.socket.connected) {
                instance.web.notification.do_warn('Impresión', 'La impresora no está preparada');
                return;
            }

            instance.web.blockUI();

            var mime_type = 'data:application/pdf;base64,';
            
            if (data instanceof Uint8Array) {
                data = mime_type + btoa(String.fromCharCode.apply(null, data));
            }

            if (!data.startsWith(mime_type)) {
                data = mime_type + data;
            }

            try {
                atob(data.replace(mime_type, ''));

                this.socket.emit('request_print', {
                    print_directly: print_directly || this.config.print_directly,
                    data: data
                });
            } catch (e) {
                instance.web.unblockUI();
                instance.webclient.crashmanager.show_message(e);
            }
        }
    });

    // Diálogo que sirve para descargar del repositorio el programa cliente
    var DownloadPrinterTools = instance.web.Widget.extend({
        template: 'DownloadPrinterTools',
        events: {
            'click .download_button': 'handle_download'
        },
        init: function (parent) {
            this._super(parent);
        },
        start: function () {
            this.renderElement();
            this.$el.appendTo($('body'));
            this.$el.on('hidden.bs.modal', this, this.on_hide);

            if (this.is_linux_platform()) {
                this.$('.download_logo_windows').css('display', 'none');
                this.$('.windows_download_button').css('display', 'none');
            }
            
            if (this.is_win_platform()) {
                this.$('.download_logo_linux').css('display', 'none');
                this.$('.linux_download_button').css('display', 'none');
            }

            this.$el.modal('show');
        },
        on_hide: function (e) {
            var self = e.data;
            self.$el.remove();
            self.destroy();
        },
        is_win_platform: function () {
            return /^Win(?:32|64)$/.test(navigator.platform);
        },
        is_linux_platform: function () {
            return /^Linux\sx86(?:|_64)$/.test(navigator.platform);
        },
        handle_download: function (e) {
            var download_url = 'https://repo.eiru.com.py/attachments/';

            if (this.is_linux_platform()) {
                download_url += '94bcc4d4-8fe7-45f9-8715-97c11fdcdd6f';
            }

            if (this.is_win_platform()) {
                download_url += '8b9b5283-c5f8-4a62-a798-2691b72936bd';
            }

            var a = document.createElement('a');
    
            a.href = download_url;
            a.target = '_parent';
            a.click();

            this.$el.modal('hide');
        }
    });

    // Icono de notificación en la barra superior del sistema
    var PrinterTopNotificator = instance.web.Widget.extend({
        template: 'PrinterTopNotificator',
        events: {
            'click .printer_status_notificator': 'handle_click'
        },
        init: function (parent) {
            this._super(parent);
        },
        handle_click: function (e) {
            e.preventDefault();

            if (local.socket_manager.socket.disconnected) {
                var widget = new DownloadPrinterTools(this);
                widget.start();
            }
        }
    });

    // Diálogo de impresión cuando la impresión directa no está disponible
    var PrinterUnavailableDialog = instance.web.Widget.extend({
        template: 'PrinterUnavailableDialog',
        events: {
            'click li': 'on_download'
        },
        init: function (parent) {
            this._super(parent);
            this.start();

            var config = local.socket_manager.config;
            this.preview_pdf = config.action_preview_pdf || false;
        },
        start: function () {
            this.defer = $.Deferred();
            this.render_widget();
        },
        render_widget: function () {
            this.renderElement();
            $('body').append(this.$el);
            
            this.$el.on('hidden.bs.modal', this, this.on_hide);

            this.$el.modal('show');
        },
        on_download: function () {
            this.download_pdf = true;
            this.$el.modal('hide');
        },
        on_hide: function (e) {
            var self = e.data;
            
            self.defer.resolve(!!self.download_pdf);

            self.destroy();
            self.$el.remove();
        },
        can_download: function () {
            return this.defer;
        }
    });

    // Diálogo de selección de la impresora que se encargará de la impresión
    var PrinterSelectionDialog = instance.web.Widget.extend({
        template: 'PrinterSelectionDialog',
        events: {
            'click li': 'on_select'
        },
        init: function (data) {
            var config = local.socket_manager.config;
            this.download_pdf = config.action_download_pdf || false;
            this.preview_pdf = config.action_preview_pdf || false;

            this.printers = data.printers || [];
            this.id = data.id;

            this.start();
        },
        start: function () {
            this.defer = $.Deferred();
            this.render_widget();
        },
        render_widget: function () {
            this.renderElement();
            $('body').append(this.$el);
            
            this.$el.on('hidden.bs.modal', this, this.on_hide);

            this.$el.modal('show');
        },
        get_selection: function () {
            return this.defer;
        },
        on_select: function (e) {
            var $el = $(e.target).closest('li');
            var name = $el.data('name');
            this.selected_printer = name;
            this.$el.modal('hide');
        },
        on_hide: function (e) {
            var self = e.data;

            self.defer.resolve({
                id: self.id,
                printer: self.selected_printer,
                print_directly: true
            });

            self.$el.remove();
        }
    });

    // Pequeño visor pdf
    var PdfViewerDialog = instance.web.Widget.extend({
        template: 'PdfViewerDialog',
        events: {
            'click a.btn': 'on_action'
        },
        init: function (parent, source, name) {
            this._super(parent);

            this.source = source;
            this.name = name;

            this.set('current_page', -1);
            this.on('change:current_page', this, this.render_page);

            if (_.isObject(this.source)) {
                this.pdf = source;
                this.render_widget();

                return;
            }

            var matchReportUrl = this.source.match(/(?:\/(?:web\/)?report(?:\/pdf)?)/);

            if (matchReportUrl) {
                var self = this;

                instance.web.blockUI();
                instance.pdfJs.getDocument(self.source).then(function (pdf) {
                    self.pdf = pdf;
                    self.render_widget();
                });

                return;
            }

            var matchMimeType = this.source.match(/data:[a-zA-Z0-9]+\/[a-zA-Z0-9-.+]+.*,/);

            if (matchMimeType) {
                if (self.source.startsWith('data:application/pdf;base64,')) {
                    self.source = {
                        data: self.source
                    };
                }
            }

            try {
                atob(this.source.data || this.source);
            } catch (e) {
                instance.webclient.crashmanager.show_message('Unknown data format');
            }

            var self = this;

            instance.web.blockUI();
            instance.pdfJs.getDocument(self.source.data || self.source).then(function (pdf) {
                self.pdf = pdf;
                self.render_widget();
            });
        },
        render_widget: function () {
            this.renderElement();

            $('body').append(this.$el);

            this.$el.on('shown.bs.modal', this, this.render_pdf);
            this.$el.on('hidden.bs.modal', this, this.on_hide);
            this.$el.modal('show');
        },
        render_pdf: function (e) {
            var self = e.data;
            
            if (!self.pdf) {
                instance.web.unblockUI();
                instance.web.notification.do_warn('No hay archivo PDF');
                return;
            }

            self.set('current_page', 1);
        },
        render_page: function () {            
            var self = this;
            var current_page = self.get('current_page');

            self.pdf.getPage(current_page).then(function (page) {
                var scale = 1;
                var viewport = page.getViewport(scale);

                var $canvas = self.$el.find('canvas');
                var context =  $canvas.get(0).getContext('2d');
                
                $canvas.attr('height', viewport.height);
                $canvas.attr('width', viewport.width);

                page.render({
                    canvasContext: context,
                    viewport: viewport
                }).then(function () {
                    self.$('.page_counter').text('Página ' + current_page + ' de ' + self.pdf.numPages);
                    instance.web.unblockUI();
                });
                
            }, function (error) {
                instance.web.unblockUI();
                instance.webclient.crashmanager.show_message(error);
            });
        },
        on_hide: function (e) {
            var self = e.data;

            if (self.pdf) {
                self.pdf.destroy();
            }

            self.destroy();
            self.$el.remove();
        },
        on_action: function (e) {
            e.preventDefault();

            var action = $(e.target).closest('a').data('action');
            this['on_' + action]();
        },
        on_next: function () {
            var current_page = this.get('current_page');
            
            if (current_page >= this.pdf.numPages) {
                return;
            }

            this.set('current_page', current_page + 1);
        },
        on_previous: function () {
            var current_page = this.get('current_page');

            if (current_page <= 1) {
                return;
            }

            this.set('current_page', current_page - 1);
        },
        on_print: function () {
            var self = this;

            if (!self.pdf) {
                return;
            }

            instance.web.blockUI();

            self.pdf.getData().then(function (data) {
                instance.web.unblockUI();

                if (local.socket_manager.socket.connected) {
                    local.socket_manager.request_print(data, true);
                    return;
                }

                instance.printJs({
                    printable: instance.pdfJs.createObjectURL(data, 'application/pdf')
                });
            });
        },
        on_download: function () {
            var self = this;

            if (!self.pdf) {
                return;
            }

            self.pdf.getMetadata().then(function (metadata) {
                self.pdf.getData().then(function (data) {
                    var url = instance.pdfJs.createObjectURL(data, 'application/pdf');
                    var a = document.createElement('a');

                    a.href = url;
                    a.download = (self.name || metadata.contentDispositionFilename || 'document') + '.pdf';
                    a.target = '_parent';
                    a.click();
                });
            });
        }
    });

    // Dispara el visor pdf
    var preview_pdf = function (owner, data, name) {
        new PdfViewerDialog(owner, data, name);
    };

    // Dispara la descarga del pdf
    var download_pdf = function (source) {
        var matchMimeType = source.match(/data:[a-zA-Z0-9]+\/[a-zA-Z0-9-.+]+.*,/);

        if (!matchMimeType) {
            instance.webclient.crashmanager.show_message('Formato no válido para la impresión');
            return;
        }
        
        if (!source.startsWith(matchMimeType[0])) {
            instance.webclient.crashmanager.show_message('Formato no válido para la impresión');
            return;
        }

        var a = document.createElement('a');

        a.href = source;
        a.download = 'documento.pdf';
        a.target = '_parent';
        a.click();
    };

    // Dispara la impresión directa o descarga pdf
    local.print = function (data) {
        var config = this.socket_manager.config;

        if (config.is_mobile) {
            download_pdf(data);
            return;
        }

        if (config.action_download_pdf) {
            download_pdf(data);
            return;
        }

        if (this.socket_manager.socket.disconnected) {
            var widget = new PrinterUnavailableDialog();

            widget.can_download().then(function (can_download) {
                if (!can_download) {
                    return;
                }

                if (config.action_preview_pdf) {
                    preview_pdf(null, data, null);
                    return;
                }

                download_pdf(data);
            });

            return;
        }

        this.socket_manager.request_print(data);
    };

    // Test
    local.TicketTestPage = instance.web.Widget.extend({
        template: 'TicketTestPage',
        events: {
            'click .ticket-test-btn': 'print_test' 
        },
        start: function () {

        },
        print_test: function (e) {
            e.preventDefault();

            var wrapper = document.createElement('div');
            wrapper.innerHTML = openerp.web.qweb.render('EiruPosTicket');
            wrapper.setAttribute('id', 'ticket_wrapper')

            $('body').append(wrapper);

            // var ticket_el = $('body').find('.eiru_pos_ticket').get(0);
            var ticket_el = document.querySelector('.eiru_pos_ticket');

            console.log(ticket_el);

            var pdf = new jsPDF({
                unit: 'mm',
                format: [290, 70]
            });

            instance.html2canvas(ticket_el, {
                canvas: pdf.canvas,
                logging: false,
                removeContainer: true,
                width: 264.5,
                height: 1016.6
            }).then(function (canvas) {
                console.log(canvas);
            });

            // var data = pdf.output('datauristring');
            // console.log(data);
            // local.print(data);
        }
    });

    instance.web.client_actions.add('printer_bridge.test_ticket', 'instance.printer_bridge.TicketTestPage');

    if (instance.web) {
        if (instance.web.UserMenu) {
            instance.web.UserMenu.include({
                do_update: function () {
                    var printer = new PrinterTopNotificator(this);
                    printer.appendTo($('.oe_systray'));
                    
                    local.socket_manager = new SocketManager();
                    local.print_directly = local.socket_manager.request_print;

                    return this._super.apply(this, arguments);
                }
            });
        }

        if (instance.web.FormView) {
            instance.web.FormView.include({
                record_saved: function (r) {
                    if (this.model === 'res.users') {
                        local.socket_manager.update_config();
                    }

                    return this._super.apply(this, arguments);
                }
            });
        }

        if (instance.web.ActionManager) {
            instance.web.ActionManager.include({
                trigger_preview_pdf: function (action) {
                    var self = this;
                    var regex = /^(?:qweb-(pdf|html)|controller)$/;

                    if (_.has(action, 'report_type') && regex.test(action.report_type)) {
                        var format = action.report_type.match(regex)[1];
                        var url = (format || action.report_file) && ('/report/' + format + '/' + action.report_name);

                        if (_.isEmpty(action.data) && _.has(action.context, 'active_ids')) {
                            url += '/' + action.context.active_ids.join(',');
                        } else {
                            url += '?options=' + encodeURIComponent(JSON.stringify(action.data)) + '&context=' + encodeURIComponent(JSON.stringify(action.context));
                        }
                        
                        preview_pdf(self, url, action.name);
                        return;
                    }
                    
                    var params = {
                        action: JSON.stringify(action),
                        token: (new Date()).getTime()
                    };
                    
                    preview_pdf(self, this.session.url('/web/report', params), action.name);
                },
                trigger_download_pdf: function (action) {
                    var regex = /^(?:qweb-(pdf|html)|controller)$/;
                    var url = null;

                    instance.web.blockUI();

                    if (_.has(action, 'report_type') && regex.test(action.report_type)) {
                        var format = action.report_type.match(regex)[1];
                        url = (format || action.report_file) && ('/report/' + format + '/' + action.report_name);

                        if (_.isEmpty(action.data) && _.has(action.context, 'active_ids')) {
                            url += '/' + action.context.active_ids.join(',');
                        } else {
                            url += '?options=' + encodeURIComponent(JSON.stringify(action.data)) + '&context=' + encodeURIComponent(JSON.stringify(action.context));
                        }
                    }
                    
                    if (!url) {
                        var params = {
                            action: JSON.stringify(action),
                            token: (new Date()).getTime()
                        };

                        url = this.session.url('/web/report', params);
                    }

                    instance.pdfJs.getDocument(url).then(function (pdf) {
                        pdf.getData().then(function (data) {
                            instance.web.unblockUI();

                            var url = instance.pdfJs.createObjectURL(data, 'application/pdf');
                            var a = document.createElement('a');
        
                            a.href = url;
                            a.download = (action.name || action.report_name) + '.pdf';
                            a.target = '_parent';
                            a.click();

                            pdf.destroy();
                        });
                    });
                },
                trigger_print_pdf: function (action) {
                    var url = '/printer_bridge/get_pdf';
                    var ctx = _.clone(action.context);

                    ctx.report_name = action.report_name;

                    instance.web.jsonRpc(url, 'call', {
                        context: ctx
                    }).done(function (result) {
                        local.socket_manager.request_print(result.data);
                    }).fail(function (e) {
                        instance.webclient.crashmanager.show_message(e);
                    });
                },
                ir_actions_report_xml: function (action, options) {
                    var self = this;
                    var config = local.socket_manager.config;

                    var eval_contexts = ([instance.session.user_context] || []).concat([action.context]);
                    action.context = instance.web.pyeval.eval('contexts', eval_contexts);

                    if (config.is_mobile) {
                        this.trigger_preview_pdf(action);
                        return;
                    }
                    
                    if (config.action_download_pdf) {
                        self.trigger_download_pdf(action);
                        return;
                    }

                    if (local.socket_manager.socket.disconnected) {
                        var widget = new PrinterUnavailableDialog(this);

                        widget.can_download().then(function (can_download) {
                            if (!can_download) {
                                return;
                            }

                            if (config.action_preview_pdf) {
                                self.trigger_preview_pdf(action);
                                return;
                            }

                            self.trigger_download_pdf(action);
                        });

                        return;
                    }

                    self.trigger_print_pdf(action);
                }
            });
        }
    }
}
